package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.Streams
import com.google.gson.GsonBuilder
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonSyntaxException
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.MinecraftServer
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.server.packs.resources.SimpleJsonResourceReloadListener
import net.minecraft.util.profiling.ProfilerFiller
import net.minecraft.world.Container
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeType
import net.minecraftforge.event.AddReloadListenerEvent
import net.minecraftforge.event.server.ServerStartedEvent
import net.minecraftforge.eventbus.api.IEventBus
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.ForgeRegistries
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.SystemTime
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.isActuallyEmpty
import ru.dbotthepony.mc.otm.registry.RegistryDelegate
import java.util.Arrays
import java.util.stream.Stream

/**
 * This manager holds json defined recipe resolvers (they process their own)
 */
object RecipeResolverManager : SimpleJsonResourceReloadListener(GsonBuilder().setPrettyPrinting().disableHtmlEscaping().create(), "otm_recipe_resolver") {
	/**
	 * because it is immutable
	 */
	data class ImmutableStack(
		val item: Item,
		val multiplier: Double = 1.0
	) {
		constructor(item: Item, count: Int) : this(item, 1.0 / count.toDouble())
		constructor(item: ItemStack) : this(item.item, item.count)
	}

	class ResolvedRecipe(
		inputs: Stream<Stream<ImmutableStack>>,
		val output: ImmutableStack,
		val transformMatterValue: (ImpreciseFraction) -> ImpreciseFraction = { it },
		val transformComplexity: (Double) -> Double = { it },

		/**
		 * if we can't resolve one of ingredient's matter value - consider output also having no matter value
		 */
		val isCritical: Boolean = true,
	) {
		val inputs: List<List<ImmutableStack>> = inputs
			.map { it.filter { it.multiplier > 0.0 }.collect(ImmutableList.toImmutableList()) }
			.filter { it.isNotEmpty() }
			.collect(ImmutableList.toImmutableList())

		constructor(
			inputs: Collection<Collection<ImmutableStack>>,
			output: ImmutableStack,
			transformMatterValue: (ImpreciseFraction) -> ImpreciseFraction = { it },
			transformComplexity: (Double) -> Double = { it },
		) : this(inputs.stream().map { it.stream() }, output, transformMatterValue, transformComplexity)
	}

	fun interface Resolver {
		/**
		 * Returned stream should be thread safe
		 */
		fun resolve(server: MinecraftServer, json: JsonObject): Stream<ResolvedRecipe>
	}

	private val delegate = RegistryDelegate<Resolver>("recipe_resolver") {
		disableSync()
		disableSaving()
	}

	private val registrar = DeferredRegister.create(delegate.key, OverdriveThatMatters.MOD_ID)
	private val LOGGER = LogManager.getLogger()

	init {
		registrar.register("simple") {
			Resolver { server, data ->
				val location = (data["recipe_type"] ?: throw JsonSyntaxException("Missing recipe type")).let { ResourceLocation.tryParse(it.asString) } ?: throw JsonSyntaxException("Invalid recipe type: ${data["recipe_type"]}")

				if (!ForgeRegistries.RECIPE_TYPES.containsKey(location)) {
					LOGGER.error("Invalid or missing recipe category: $location!")
					return@Resolver Stream.empty()
				}

				val findRecipeType = ForgeRegistries.RECIPE_TYPES.getValue(location) as RecipeType<Recipe<Container>>? ?: throw ConcurrentModificationException()

				val isCritical = data["is_critical"]?.asBoolean ?: true
				val ignoreDamageables = data["ignore_damageables"]?.asBoolean ?: false

				server.recipeManager.byType(findRecipeType).values.parallelStream()
					.filter { !it.isIncomplete && !it.ingredients.stream().anyMatch { it.isActuallyEmpty } } // get rid of invalid recipes, second "isActuallyEmpty" is required because we do care about ingredients being "missing"
					.filter {
						!ignoreDamageables || it.ingredients.stream().flatMap { Arrays.stream(it.items) }.noneMatch { it.isDamageableItem }
					}
					.map {
						ResolvedRecipe(
							it.ingredients.stream().map { Arrays.stream(it.items).map { ImmutableStack(it) } },
							ImmutableStack(it.resultItem),
							isCritical = isCritical
						)
					}
					.filter { it.inputs.isNotEmpty() }
			}
		}
	}

	fun initialize(bus: IEventBus) {
		bus.addListener(delegate::build)
		registrar.register(bus)
	}

	fun reloadEvent(event: AddReloadListenerEvent) {
		event.addListener(this)
	}

	fun onServerStarted(event: ServerStartedEvent) {
		resolve(event.server)
	}

	@JvmStatic val RESOLVERS get() = delegate.get()
	@JvmStatic val RESOLVERS_KEY get() = delegate.key

	/**
	 * it exists solely for fail-fast to find bugs, do not rely on this property
	 */
	var ready = false
		private set

	var resolved = false
		private set

	private var foundResolvers: Map<ResourceLocation, Pair<Resolver, JsonObject>> = ImmutableMap.of()

	override fun prepare(
		p_10771_: ResourceManager,
		p_10772_: ProfilerFiller
	): MutableMap<ResourceLocation, JsonElement> {
		ready = false
		resolved = false
		// determinedValues.clear() // can't clear here, it is executed on separate thread
		return super.prepare(p_10771_, p_10772_)
	}

	override fun apply(
		map: Map<ResourceLocation, JsonElement>,
		resourceManager: ResourceManager,
		profilerFiller: ProfilerFiller
	) {
		ready = false
		resolved = false
		determinedValues.clear()

		val builder = ImmutableMap.Builder<ResourceLocation, Pair<Resolver, JsonObject>>()

		for ((key, json) in map) {
			if (json !is JsonObject) {
				throw JsonParseException("$key is not a json object")
			}

			val location = (json["type"] ?: throw JsonSyntaxException("Missing resolver type")).let { ResourceLocation.tryParse(it.asString) } ?: throw JsonSyntaxException("Invalid resolver type: ${json["type"]}")

			if (!RESOLVERS.containsKey(location)) {
				throw JsonParseException("Resolver type $location does not exist (in $key)")
			}

			val resolver = RESOLVERS.getValue(location) ?: throw ConcurrentModificationException()
			builder.put(key, resolver to json)
		}

		foundResolvers = builder.build()

		ready = true
	}

	private data class Accumulator(
		val input2Recipes: Reference2ObjectOpenHashMap<Item, ReferenceOpenHashSet<ResolvedRecipe>> = Reference2ObjectOpenHashMap(),
		val output2Recipes: Reference2ObjectOpenHashMap<Item, ReferenceOpenHashSet<ResolvedRecipe>> = Reference2ObjectOpenHashMap(),
	) {
		fun combine(other: Accumulator) {
			input2Recipes.putAll(other.input2Recipes)
			output2Recipes.putAll(other.output2Recipes)
		}

		fun add(recipe: ResolvedRecipe) {
			output2Recipes.computeIfAbsent(recipe.output.item, Reference2ObjectFunction { ReferenceOpenHashSet() }).add(recipe)

			for (input1 in recipe.inputs) {
				for (input in input1) {
					input2Recipes.computeIfAbsent(input.item, Reference2ObjectFunction { ReferenceOpenHashSet() }).add(recipe)
				}
			}
		}
	}

	fun getDetermined(index: Item): MatterManager.IMatterValue {
		return determinedValues[index] ?: MatterManager.IMatterValue.Companion
	}

	private var input2Recipes: Map<Item, Collection<ResolvedRecipe>> = mapOf()
	private var output2Recipes: Map<Item, Collection<ResolvedRecipe>> = mapOf()

	private val determinedValues = Reference2ObjectOpenHashMap<Item, MatterManager.IMatterValue>()
	private val seenItems = ArrayDeque<Item>()

	private data class Result(val type: Type, val value: MatterManager.IMatterValue? = null) {
		constructor(value: MatterManager.IMatterValue) : this(Type.RESOLVED, value)

		val isSkipped get() = type === Type.SKIPPED
		val isMissing get() = type === Type.MISSING

		enum class Type {
			RESOLVED, SKIPPED, MISSING
		}

		companion object {
			// recursive recipes should be ignored until we resolve everything else
			val SKIPPED = Result(Type.SKIPPED)

			// missing matter values are fatal, whole recipe chain is then considered defunct
			val MISSING = Result(Type.MISSING)
		}
	}

	private var changes = false
	private var cachedIterationResults = Reference2ObjectOpenHashMap<Item, Result>()

	private fun doDetermineValue(item: Item): Result {
		var minimalMatter: ImpreciseFraction? = null
		var minimalComplexity: Double? = null

		recipesLoop@ for (recipe in output2Recipes[item] ?: return Result.MISSING) {
			if (recipe.inputs.isEmpty()) {
				// TODO: should we ignore empty recipes?
				continue
			}

			var accumulatedMatter: ImpreciseFraction? = null
			var accumulatedComplexity: Double? = null

			inputsLoop@ for (inputs in recipe.inputs) {
				var minimal: MatterManager.IMatterValue? = null
				var minimalMultiplier = 0.0

				for (input in inputs) {
					val ivalue = determineValue(input.item)

					if (ivalue.isMissing) {
						if (recipe.isCritical) {
							return Result.MISSING
						} else {
							continue@recipesLoop
						}
					} else if (ivalue.isSkipped) {
						return Result.SKIPPED
					}

					if (minimal == null || minimal > ivalue.value!!) {
						minimal = ivalue.value
						minimalMultiplier = input.multiplier
					}
				}

				if (minimal == null || !minimal.hasMatterValue)
					return Result.MISSING

				if (accumulatedMatter == null || accumulatedComplexity == null) {
					accumulatedMatter = minimal.matter * minimalMultiplier
					accumulatedComplexity = minimal.complexity * minimalMultiplier
				} else {
					accumulatedMatter += minimal.matter * minimalMultiplier
					accumulatedComplexity += minimal.complexity * minimalMultiplier
				}
			}

			if (accumulatedMatter == null || accumulatedComplexity == null) {
				// TODO: should we ignore empty recipes?
				continue
			}

			if (!accumulatedMatter.isPositive || accumulatedComplexity <= 0.0) {
				return Result.MISSING
			}

			if (minimalMatter == null || minimalMatter > accumulatedMatter) {
				minimalMatter = recipe.transformMatterValue(accumulatedMatter * recipe.output.multiplier)
				minimalComplexity = recipe.transformComplexity(accumulatedComplexity * recipe.output.multiplier)
			}
		}

		if (minimalMatter == null || minimalComplexity == null) {
			return Result.MISSING
		}

		val result = MatterManager.MatterValue(minimalMatter, minimalComplexity)

		changes = true
		determinedValues[item] = result

		return Result(result)
	}

	private fun determineValue(item: Item): Result {
		val getResult = cachedIterationResults[item]

		if (getResult != null) {
			return getResult
		}

		var value: MatterManager.IMatterValue? = MatterManager.getDirect(item)

		if (value?.hasMatterValue == true) {
			return Result(value)
		}

		value = determinedValues[item]

		if (value?.hasMatterValue == true) {
			return Result(value)
		}

		if (item in seenItems) {
			return Result.SKIPPED
		}

		seenItems.addLast(item)

		try {
			val result = doDetermineValue(item)
			cachedIterationResults[item] = result
			return result
		} finally {
			seenItems.removeLast()
		}
	}

	private fun resolve(server: MinecraftServer) {
		if (resolved) {
			return
		}

		check(ready) { "RecipeResolverManager is not ready to resolve matter values" }
		check(MatterManager.canCompute) { "MatteryManager is not ready to compute matter values" }

		var time = SystemTime()
		LOGGER.info("Finding recipes...")

		determinedValues.clear()
		seenItems.clear()

		val stream = Streams.concat(*foundResolvers.map {
			try {
				it.value.first.resolve(server, it.value.second).parallel()
			} catch(err: Throwable) {
				LOGGER.fatal("Recipe resolver ${it.key} experienced internal error", err)
				throw RuntimeException("Recipe resolver ${it.key} experienced internal error", err)
			}
		}.toTypedArray())

		val (input2Recipes, output2Recipes) = stream.collect(::Accumulator, Accumulator::add, Accumulator::combine)

		LOGGER.info("Finding recipes took ${time.millis}ms, involving ${input2Recipes.keys.size} unique inputs and ${output2Recipes.keys.size} unique outputs")
		time = SystemTime()
		LOGGER.info("Resolving recipes...")

		this.input2Recipes = input2Recipes
		this.output2Recipes = output2Recipes

		changes = true
		var i = 0
		var ops = 0

		val toDetermine = ReferenceLinkedOpenHashSet<Item>()
		toDetermine.addAll(input2Recipes.keys)
		toDetermine.addAll(output2Recipes.keys)

		while (changes) {
			ops += cachedIterationResults.size
			cachedIterationResults = Reference2ObjectOpenHashMap()
			changes = false
			i++

			val iterator = toDetermine.iterator()

			for (value in iterator) {
				if (determineValue(value).value?.hasMatterValue == true) {
					iterator.remove()
				}
			}
		}

		LOGGER.info("Resolving recipes took ${time.millis}ms in $i iterations with ~$ops operations, determined ${determinedValues.size} matter values")
	}
}
